# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow/contrib/tensor_forest/proto/fertile_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorflow.contrib.decision_trees.proto import generic_tree_model_pb2 as tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow/contrib/tensor_forest/proto/fertile_stats.proto',
  package='tensorflow.tensorforest',
  syntax='proto3',
  serialized_options=_b('\370\001\001'),
  serialized_pb=_b('\n:tensorflow/contrib/tensor_forest/proto/fertile_stats.proto\x12\x17tensorflow.tensorforest\x1a@tensorflow/contrib/decision_trees/proto/generic_tree_model.proto\"J\n\x0c\x46\x65rtileStats\x12:\n\x0cnode_to_slot\x18\x01 \x03(\x0b\x32$.tensorflow.tensorforest.FertileSlot\"\x1b\n\tGiniStats\x12\x0e\n\x06square\x18\x02 \x01(\x02\"\xd2\x04\n\x08LeafStat\x12\x12\n\nweight_sum\x18\x03 \x01(\x02\x12[\n\x0e\x63lassification\x18\x01 \x01(\x0b\x32\x41.tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStatsH\x00\x12S\n\nregression\x18\x02 \x01(\x0b\x32=.tensorflow.tensorforest.LeafStat.LeastSquaresRegressionStatsH\x00\x1a\xda\x01\n\x1fGiniImpurityClassificationStats\x12\x39\n\x0c\x64\x65nse_counts\x18\x01 \x01(\x0b\x32!.tensorflow.decision_trees.VectorH\x00\x12@\n\rsparse_counts\x18\x02 \x01(\x0b\x32\'.tensorflow.decision_trees.SparseVectorH\x00\x12\x30\n\x04gini\x18\x03 \x01(\x0b\x32\".tensorflow.tensorforest.GiniStatsB\x08\n\x06\x63ounts\x1a\x95\x01\n\x1bLeastSquaresRegressionStats\x12\x36\n\x0bmean_output\x18\x01 \x01(\x0b\x32!.tensorflow.decision_trees.Vector\x12>\n\x13mean_output_squares\x18\x02 \x01(\x0b\x32!.tensorflow.decision_trees.VectorB\x0b\n\tleaf_stat\"\xe2\x01\n\x0b\x46\x65rtileSlot\x12\x35\n\nleaf_stats\x18\x04 \x01(\x0b\x32!.tensorflow.tensorforest.LeafStat\x12;\n\ncandidates\x18\x01 \x03(\x0b\x32\'.tensorflow.tensorforest.SplitCandidate\x12?\n\x14post_init_leaf_stats\x18\x06 \x01(\x0b\x32!.tensorflow.tensorforest.LeafStat\x12\x0f\n\x07node_id\x18\x05 \x01(\x05\x12\r\n\x05\x64\x65pth\x18\x07 \x01(\x05\"\xc8\x01\n\x0eSplitCandidate\x12\x34\n\x05split\x18\x01 \x01(\x0b\x32%.tensorflow.decision_trees.BinaryNode\x12\x35\n\nleft_stats\x18\x04 \x01(\x0b\x32!.tensorflow.tensorforest.LeafStat\x12\x36\n\x0bright_stats\x18\x05 \x01(\x0b\x32!.tensorflow.tensorforest.LeafStat\x12\x11\n\tunique_id\x18\x06 \x01(\t\"F\n\x08TreePath\x12:\n\rnodes_visited\x18\x01 \x03(\x0b\x32#.tensorflow.decision_trees.TreeNodeB\x03\xf8\x01\x01\x62\x06proto3')
  ,
  dependencies=[tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2.DESCRIPTOR,])




_FERTILESTATS = _descriptor.Descriptor(
  name='FertileStats',
  full_name='tensorflow.tensorforest.FertileStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_to_slot', full_name='tensorflow.tensorforest.FertileStats.node_to_slot', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=153,
  serialized_end=227,
)


_GINISTATS = _descriptor.Descriptor(
  name='GiniStats',
  full_name='tensorflow.tensorforest.GiniStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='square', full_name='tensorflow.tensorforest.GiniStats.square', index=0,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=229,
  serialized_end=256,
)


_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS = _descriptor.Descriptor(
  name='GiniImpurityClassificationStats',
  full_name='tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dense_counts', full_name='tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStats.dense_counts', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sparse_counts', full_name='tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStats.sparse_counts', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gini', full_name='tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStats.gini', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='counts', full_name='tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStats.counts',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=470,
  serialized_end=688,
)

_LEAFSTAT_LEASTSQUARESREGRESSIONSTATS = _descriptor.Descriptor(
  name='LeastSquaresRegressionStats',
  full_name='tensorflow.tensorforest.LeafStat.LeastSquaresRegressionStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mean_output', full_name='tensorflow.tensorforest.LeafStat.LeastSquaresRegressionStats.mean_output', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_output_squares', full_name='tensorflow.tensorforest.LeafStat.LeastSquaresRegressionStats.mean_output_squares', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=691,
  serialized_end=840,
)

_LEAFSTAT = _descriptor.Descriptor(
  name='LeafStat',
  full_name='tensorflow.tensorforest.LeafStat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='weight_sum', full_name='tensorflow.tensorforest.LeafStat.weight_sum', index=0,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='classification', full_name='tensorflow.tensorforest.LeafStat.classification', index=1,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='regression', full_name='tensorflow.tensorforest.LeafStat.regression', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS, _LEAFSTAT_LEASTSQUARESREGRESSIONSTATS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='leaf_stat', full_name='tensorflow.tensorforest.LeafStat.leaf_stat',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=259,
  serialized_end=853,
)


_FERTILESLOT = _descriptor.Descriptor(
  name='FertileSlot',
  full_name='tensorflow.tensorforest.FertileSlot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='leaf_stats', full_name='tensorflow.tensorforest.FertileSlot.leaf_stats', index=0,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='candidates', full_name='tensorflow.tensorforest.FertileSlot.candidates', index=1,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='post_init_leaf_stats', full_name='tensorflow.tensorforest.FertileSlot.post_init_leaf_stats', index=2,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node_id', full_name='tensorflow.tensorforest.FertileSlot.node_id', index=3,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='depth', full_name='tensorflow.tensorforest.FertileSlot.depth', index=4,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=856,
  serialized_end=1082,
)


_SPLITCANDIDATE = _descriptor.Descriptor(
  name='SplitCandidate',
  full_name='tensorflow.tensorforest.SplitCandidate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='split', full_name='tensorflow.tensorforest.SplitCandidate.split', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='left_stats', full_name='tensorflow.tensorforest.SplitCandidate.left_stats', index=1,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='right_stats', full_name='tensorflow.tensorforest.SplitCandidate.right_stats', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unique_id', full_name='tensorflow.tensorforest.SplitCandidate.unique_id', index=3,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1085,
  serialized_end=1285,
)


_TREEPATH = _descriptor.Descriptor(
  name='TreePath',
  full_name='tensorflow.tensorforest.TreePath',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nodes_visited', full_name='tensorflow.tensorforest.TreePath.nodes_visited', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1287,
  serialized_end=1357,
)

_FERTILESTATS.fields_by_name['node_to_slot'].message_type = _FERTILESLOT
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['dense_counts'].message_type = tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2._VECTOR
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['sparse_counts'].message_type = tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2._SPARSEVECTOR
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['gini'].message_type = _GINISTATS
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.containing_type = _LEAFSTAT
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.oneofs_by_name['counts'].fields.append(
  _LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['dense_counts'])
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['dense_counts'].containing_oneof = _LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.oneofs_by_name['counts']
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.oneofs_by_name['counts'].fields.append(
  _LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['sparse_counts'])
_LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.fields_by_name['sparse_counts'].containing_oneof = _LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS.oneofs_by_name['counts']
_LEAFSTAT_LEASTSQUARESREGRESSIONSTATS.fields_by_name['mean_output'].message_type = tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2._VECTOR
_LEAFSTAT_LEASTSQUARESREGRESSIONSTATS.fields_by_name['mean_output_squares'].message_type = tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2._VECTOR
_LEAFSTAT_LEASTSQUARESREGRESSIONSTATS.containing_type = _LEAFSTAT
_LEAFSTAT.fields_by_name['classification'].message_type = _LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS
_LEAFSTAT.fields_by_name['regression'].message_type = _LEAFSTAT_LEASTSQUARESREGRESSIONSTATS
_LEAFSTAT.oneofs_by_name['leaf_stat'].fields.append(
  _LEAFSTAT.fields_by_name['classification'])
_LEAFSTAT.fields_by_name['classification'].containing_oneof = _LEAFSTAT.oneofs_by_name['leaf_stat']
_LEAFSTAT.oneofs_by_name['leaf_stat'].fields.append(
  _LEAFSTAT.fields_by_name['regression'])
_LEAFSTAT.fields_by_name['regression'].containing_oneof = _LEAFSTAT.oneofs_by_name['leaf_stat']
_FERTILESLOT.fields_by_name['leaf_stats'].message_type = _LEAFSTAT
_FERTILESLOT.fields_by_name['candidates'].message_type = _SPLITCANDIDATE
_FERTILESLOT.fields_by_name['post_init_leaf_stats'].message_type = _LEAFSTAT
_SPLITCANDIDATE.fields_by_name['split'].message_type = tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2._BINARYNODE
_SPLITCANDIDATE.fields_by_name['left_stats'].message_type = _LEAFSTAT
_SPLITCANDIDATE.fields_by_name['right_stats'].message_type = _LEAFSTAT
_TREEPATH.fields_by_name['nodes_visited'].message_type = tensorflow_dot_contrib_dot_decision__trees_dot_proto_dot_generic__tree__model__pb2._TREENODE
DESCRIPTOR.message_types_by_name['FertileStats'] = _FERTILESTATS
DESCRIPTOR.message_types_by_name['GiniStats'] = _GINISTATS
DESCRIPTOR.message_types_by_name['LeafStat'] = _LEAFSTAT
DESCRIPTOR.message_types_by_name['FertileSlot'] = _FERTILESLOT
DESCRIPTOR.message_types_by_name['SplitCandidate'] = _SPLITCANDIDATE
DESCRIPTOR.message_types_by_name['TreePath'] = _TREEPATH
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FertileStats = _reflection.GeneratedProtocolMessageType('FertileStats', (_message.Message,), {
  'DESCRIPTOR' : _FERTILESTATS,
  '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.FertileStats)
  })
_sym_db.RegisterMessage(FertileStats)

GiniStats = _reflection.GeneratedProtocolMessageType('GiniStats', (_message.Message,), {
  'DESCRIPTOR' : _GINISTATS,
  '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.GiniStats)
  })
_sym_db.RegisterMessage(GiniStats)

LeafStat = _reflection.GeneratedProtocolMessageType('LeafStat', (_message.Message,), {

  'GiniImpurityClassificationStats' : _reflection.GeneratedProtocolMessageType('GiniImpurityClassificationStats', (_message.Message,), {
    'DESCRIPTOR' : _LEAFSTAT_GINIIMPURITYCLASSIFICATIONSTATS,
    '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.LeafStat.GiniImpurityClassificationStats)
    })
  ,

  'LeastSquaresRegressionStats' : _reflection.GeneratedProtocolMessageType('LeastSquaresRegressionStats', (_message.Message,), {
    'DESCRIPTOR' : _LEAFSTAT_LEASTSQUARESREGRESSIONSTATS,
    '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.LeafStat.LeastSquaresRegressionStats)
    })
  ,
  'DESCRIPTOR' : _LEAFSTAT,
  '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.LeafStat)
  })
_sym_db.RegisterMessage(LeafStat)
_sym_db.RegisterMessage(LeafStat.GiniImpurityClassificationStats)
_sym_db.RegisterMessage(LeafStat.LeastSquaresRegressionStats)

FertileSlot = _reflection.GeneratedProtocolMessageType('FertileSlot', (_message.Message,), {
  'DESCRIPTOR' : _FERTILESLOT,
  '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.FertileSlot)
  })
_sym_db.RegisterMessage(FertileSlot)

SplitCandidate = _reflection.GeneratedProtocolMessageType('SplitCandidate', (_message.Message,), {
  'DESCRIPTOR' : _SPLITCANDIDATE,
  '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.SplitCandidate)
  })
_sym_db.RegisterMessage(SplitCandidate)

TreePath = _reflection.GeneratedProtocolMessageType('TreePath', (_message.Message,), {
  'DESCRIPTOR' : _TREEPATH,
  '__module__' : 'tensorflow.contrib.tensor_forest.proto.fertile_stats_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.tensorforest.TreePath)
  })
_sym_db.RegisterMessage(TreePath)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""AutomaticControlDependencies and related functionality."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.eager import context
from tensorflow.python.framework import dtypes as dtypes_module
from tensorflow.python.framework import ops
from tensorflow.python.framework import sparse_tensor
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import control_flow_util
from tensorflow.python.ops import tensor_array_ops
from tensorflow.python.util import nest
from tensorflow.python.util import object_identity
from tensorflow.python.util import tf_decorator

# LINT.IfChange
# Op types that should not run in program order, e.g. because they need to run
# asynchronously to avoid deadlock.
ASYNC_STATEFUL_OPS = [
    "CollectiveGather",
    "CollectiveReduce",
    "CollectiveBcastSend",
    "CollectiveBcastRecv",
    "NcclAllReduce",
]

LEGACY_RANDOM_OPS = [
    # These may be used in variable initializers -- thus their execution should
    # not be dependent on other stateful operations.  This is because although
    # according to program order, tf.Variables may be created in sequence,
    # their initialization happens outside of the program order (specifically,
    # in graph mode their initialization happens by calling a grouped
    # initializer operation or in eager mode, where initialization is lifted
    # out of the tf.function and executed the first time the function is
    # executed).
    #
    # Unless there is a specific dependency between the initializers
    # themselves (e.g. one initializer depends on a Variable whose value depends
    # on another initializer), the initialization can happen in any order so
    # long as it's before the associated Variable read operations.
    #
    # Note that in general the randomness of legacy random operations is only
    # guaranteed by providing a graph-level and op-level seed (and ordering of
    # the same op across multiple iterations of a while_loop is specifically not
    # guaranteed; see the discussion below).
    #
    # There is a possible race condition inside while_loop where the same
    # random OpKernel instantiation is reused across multiple steps
    # of the loop.  Since legacy Random OpKernels have an internal rng state,
    # automatic dependency tracking across loop steps would likely
    # fix this race; and for that case this blacklist is problematic.
    # However, since automatic dependency tracking inside while loops is not
    # currently supported, and there are no other examples of OpKernel reuse
    # (each OpKernel is associated with a unique op in graph mode),
    # this blacklist has no effect on the aforementioned behavior.
    #
    # TODO(ebrevdo,skyewm): Modify the check against this blacklist to
    # only occur when the op is inside a "variable initialization scope"; and
    # add proper autodeps inside while_loops that respects this updated check.
    "RandomUniform",
    "RandomUniformInt",
    "RandomStandardNormal",
    "ParameterizedTruncatedNormal",
    "TruncatedNormal",
    "RandomShuffle",
    "Multinomial",
    "RandomGamma",
    "RandomGammaGrad",
    "RandomPoisson",
    "RandomPoissonV2",
]
# LINT.ThenChange(//tensorflow/core/grappler/optimizers/function_optimizer.cc)

_ALL_BLACKLISTED_OPS = set(ASYNC_STATEFUL_OPS) | set(LEGACY_RANDOM_OPS)


def op_is_stateful(op):
  # pylint: disable=protected-access
  return op._is_stateful and op.type not in _ALL_BLACKLISTED_OPS


class AutomaticControlDependencies(object):
  """Context manager to automatically add control dependencies.

  Code under this context manager will act as if a sensible set of control
  dependencies were present. More specifically:
    1. All stateful ops in the scope will execute (with the exception of ops in
       ASYNC_STATEFUL_OPS and LEGACY_RANDOM_OPS)
    2. Stateful ops which modify the same resource will execute in program order

  Note: creating variables in an automatic control dependencies context is not
  supported (the value of the variables will never change as they will keep
  getting reinitialized).

  NOT THREAD SAFE
  """

  def __init__(self):
    self._returned_tensors = object_identity.ObjectIdentitySet()
    self.ops_which_must_run = set()

  def mark_as_return(self, tensor):
    """Acts like identity but marks the `Tensor` as a return value.

    This will possibly return a copy of the `Tensor`. Usage:

    ```
      with AutomaticControlDependencies() as a:
       ...
       t = a.mark_as_return(t)
      _ = ...(t...)  # i.e. it's safe to use t here
    ```

    Args:
      tensor: the `Tensor` to be marked

    Returns:
      a copy of the `Tensor`.
    """
    if isinstance(tensor, ops.IndexedSlices):
      values = array_ops.identity(tensor.values)
      indices = array_ops.identity(tensor.indices)
      self._returned_tensors.add(indices)
      self._returned_tensors.add(values)
      return ops.IndexedSlices(values, indices, dense_shape=tensor.dense_shape)
    elif isinstance(tensor, sparse_tensor.SparseTensor):
      values = array_ops.identity(tensor.values)
      indices = array_ops.identity(tensor.indices)
      self._returned_tensors.add(indices)
      self._returned_tensors.add(values)
      return sparse_tensor.SparseTensor(
          indices, values, dense_shape=tensor.dense_shape)
    elif isinstance(tensor, tensor_array_ops.TensorArray):
      flow = array_ops.identity(tensor.flow)
      self._returned_tensors.add(flow)
      return tensor_array_ops.build_ta_with_new_flow(tensor, flow)
    # We want to make the return values depend on the stateful operations, but
    # we don't want to introduce a cycle, so we make the return value the result
    # of a new identity operation that the stateful operations definitely don't
    # depend on.
    tensor = array_ops.identity(tensor)
    self._returned_tensors.add(tensor)
    return tensor

  def __enter__(self):
    if context.executing_eagerly():
      return self
    # This code assumes no other thread is adding ops to the graph while
    # we're adding ops to the graph.
    # TODO(apassos): Fix this by locking the graph or using a temporary
    # graph (but that would mess up devices and collections at least,
    # probably other things as well).
    self._graph = ops.get_default_graph()
    self._graph._add_control_dependencies = True  # pylint: disable=protected-access
    self._n_operations = len(self._graph.get_operations())
    return self

  def _process_switch(self, switch_op, ops_which_must_run,
                      last_op_using_resource_tensor, merge_for_resource):
    """Processes a switch node for a resource input.

    When tensorflow creates a cond, it creates a control flow context for each
    branch of the cond. Each external tensor accessed by that branch is routed
    through a switch op, which gets created in the graph _after_ the op which
    uses that tensor get created.

    If the resource comes from another switch op we process that one first.

    _process_switch creates a corresponding merge node for the switch node. This
    merge node is added to the outer control flow context of the switch
    node. We also ensure that:

      1. The switch node executes after the previous op which used the resource
         tensor

      2. Any op which uses a resource output of the switch node executes before
         the merge for the switch node.

      3. The next op which uses the input resource to the switch node (which
         might be another switch node for the other branch of the conditional)
         will execute after the merge node is done.

      4. The merge node is marked as must_run so it will run even if no
         subsequent operation uses the resource.

    Args:
      switch_op: the switch op to be processed
      ops_which_must_run: the set of ops which must run
      last_op_using_resource_tensor: map from resource tensor to last op using
        it
      merge_for_resource: map from resource tensor to merge which must follow
        all usages of it.
    """
    inp = switch_op.inputs[0]
    input_id = ops.tensor_id(inp)
    if inp.dtype == dtypes_module.resource and inp.op.type == "Switch":
      self._process_switch(inp.op, ops_which_must_run,
                           last_op_using_resource_tensor, merge_for_resource)
    output = switch_op.outputs[0]
    output_id = ops.tensor_id(output)
    if output_id in merge_for_resource:
      return
    new_merge = control_flow_ops.merge(switch_op.outputs,
                                       name="artificial_merge")
    new_merge[0].op._control_flow_context = (  # pylint: disable=protected-access
        switch_op._control_flow_context.outer_context)  # pylint: disable=protected-access
    # Ensures the merge always runs
    ops_which_must_run.add(new_merge[0].op)
    if input_id in last_op_using_resource_tensor:
      # Ensures the switch executes after the previous op using the resource.
      switch_op._add_control_input(last_op_using_resource_tensor[input_id])  # pylint: disable=protected-access
    # Ensure the next op outside the cond happens after the merge.
    last_op_using_resource_tensor[input_id] = new_merge[0].op
    if input_id in merge_for_resource:
      merge_for_resource[input_id]._add_control_input(new_merge[0].op)  # pylint: disable=protected-access
    for o in switch_op.outputs:
      # Ensures the merge will execute after all ops inside the cond
      merge_for_resource[ops.tensor_id(o)] = new_merge[0].op

  def __exit__(self, unused_type, unused_value, unused_traceback):
    if context.executing_eagerly():
      return

    if self._graph is not ops.get_default_graph():
      raise RuntimeError(
          "Graph changed while trying to add control dependencies.")

    # pylint: disable=protected-access
    if hasattr(self._graph, "outer_graph"):
      outer_val = self._graph.outer_graph._add_control_dependencies
      self._graph._add_control_dependencies = outer_val
    else:
      self._graph._add_control_dependencies = False
    # pylint: enable=protected-access

    # map from resource tensor to the last op which used it
    last_op_using_resource_tensor = {}
    # set of conditional and loop exits
    ops_which_must_run = set()
    # merge which must depend on ops which use this resource
    merge_for_resource = {}

    new_operations = self._graph.get_operations()[self._n_operations:]

    # Ensures that uses of resource tensors get serialized properly and all
    # execute. This is done by keeping a map from resource tensor to the last op
    # in graph-construction order which used it (last_op_using_resource_tensor).
    #
    # Conditionals are written in TensorFlow such that every external tensor
    # accessed in the conditional goes through a switch op and every return
    # tensor (it's guaranteed that there will be at least one) goes through a
    # merge op.
    #
    # To handle conditionals, switches are handled in a special way (see
    # comments for _process_switch). Merge nodes created by TF's conditional
    # logic (as opposed to by _process_switch) are forced to run and also get a
    # control dependency added to them to ensure all stateful ops inside their
    # control flow context run.
    #
    # We also ensure that if an op is using a resource output by a switch node
    # (that is, a resource tensor for which there's a value in
    # merge_for_resource) this op will run before the merge for that resource.
    #
    # We try to add control inputs to nodes respecting their control flow
    # contexts to avoid dead nodes propagating everywhere and leading to
    # "retval[0] doesn't have value" errors. If a node gets a control dependency
    # on a dead node (i.e. a note from an untaken control flow branch) that node
    # will be marked as dead unless it's a merge node.
    #
    # TODO(apassos): serialize non-resource-taking stateful ops as well, and
    # test that it works. Support while loops. Support init_scope escaping from
    # this.
    for op in new_operations:
      # TODO(apassos) make this code safely support while loops.
      if control_flow_util.IsInWhileLoop(op):
        continue
      control_inputs = set()
      # Ensure stateful ops run
      if (op.type not in self._graph._registered_ops  # pylint: disable=protected-access
          or op_is_stateful(op)):
        ops_which_must_run.add(op)
      # Ignore switches (they're handled separately)
      if op.type == "Switch" and op.inputs[0].dtype == dtypes_module.resource:
        continue
      # Make merges trigger all other computation which must run
      if op.type == "Merge":
        for o in ops_which_must_run:
          op._add_control_input(o)  # pylint: disable=protected-access
          for inp in o.inputs:
            input_id = ops.tensor_id(inp)
            if input_id in last_op_using_resource_tensor:
              last_op_using_resource_tensor[input_id] = op
        ops_which_must_run = set([op])
        continue

      resource_inputs = set()
      # Check for any resource inputs. If we find any, we update control_inputs
      # and last_op_using_resource_tensor.
      for inp in op.inputs:
        if inp.dtype != dtypes_module.resource:
          continue

        input_id = ops.tensor_id(inp)

        # If the op receives the same resource tensor twice as an input, we skip
        # to avoid the op getting a control dependency on itself.
        if input_id in resource_inputs:
          continue

        resource_inputs.add(input_id)
        # Deal with switches, finally.
        if inp.op.type == "Switch":
          self._process_switch(inp.op, ops_which_must_run,
                               last_op_using_resource_tensor,
                               merge_for_resource)
        # Ensure uses of resources are serialized
        if input_id in last_op_using_resource_tensor:
          if (last_op_using_resource_tensor[input_id]._control_flow_context  # pylint: disable=protected-access
              is op._control_flow_context):  # pylint: disable=protected-access
            control_inputs.add(last_op_using_resource_tensor[input_id])
        # Ensure merges happen after the closing of a cond block
        if input_id in merge_for_resource:
          merge_for_resource[input_id]._add_control_input(op)  # pylint: disable=protected-access
        last_op_using_resource_tensor[input_id] = op

      if (op_is_stateful(op) and not resource_inputs
          and op._control_flow_context is None):  # pylint: disable=protected-access
        if None in last_op_using_resource_tensor:
          op._add_control_input(last_op_using_resource_tensor[None])  # pylint: disable=protected-access
        last_op_using_resource_tensor[None] = op
      control_inputs = [c for c in control_inputs
                        if c._control_flow_context is op._control_flow_context]  # pylint: disable=protected-access
      op._add_control_inputs(control_inputs)  # pylint: disable=protected-access

    # Ensure all ops which must run do run
    self.ops_which_must_run.update(ops_which_must_run)
    for r in nest.flatten(list(self._returned_tensors), expand_composites=True):
      if self.ops_which_must_run:
        r.op._add_control_inputs(  # pylint: disable=protected-access
            [o for o in self.ops_which_must_run
             if o._control_flow_context is r.op._control_flow_context])  # pylint: disable=protected-access


def automatic_control_dependencies(f):
  """Wraps f to automatically insert control dependencies.

  The inserted dependencies ensure that:
    1. All stateful ops in f run when the result of f runs
    2. Updates to the same resources happen in order.

  Args:
    f: the function to be wrapped.

  Returns:
    The wrapped function.
  """

  def wrapper(*args, **kwargs):
    with AutomaticControlDependencies() as a:
      result = f(*args, **kwargs)
      result_flat = [a.mark_as_return(t) for t in nest.flatten(result)]
      return nest.pack_sequence_as(result, result_flat)

  return tf_decorator.make_decorator(f, wrapper)



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains Gradient functions for image ops."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import gen_image_ops
from tensorflow.python.ops import math_ops


@ops.RegisterGradient("ResizeNearestNeighbor")
def _ResizeNearestNeighborGrad(op, grad):
  """The derivatives for nearest neighbor resizing.

  Args:
    op: The ResizeNearestNeighbor op.
    grad: The tensor representing the gradient w.r.t. the output.

  Returns:
    The gradients w.r.t. the input and the output.
  """
  image = op.inputs[0]
  if image.get_shape()[1:3].is_fully_defined():
    image_shape = image.get_shape()[1:3]
  else:
    image_shape = array_ops.shape(image)[1:3]

  grads = gen_image_ops.resize_nearest_neighbor_grad(
      grad,
      image_shape,
      align_corners=op.get_attr("align_corners"),
      half_pixel_centers=op.get_attr("half_pixel_centers"))
  return [grads, None]


@ops.RegisterGradient("ResizeBilinear")
def _ResizeBilinearGrad(op, grad):
  """The derivatives for bilinear resizing.

  Args:
    op: The ResizeBilinear op.
    grad: The tensor representing the gradient w.r.t. the output.

  Returns:
    The gradients w.r.t. the input.
  """
  grad0 = gen_image_ops.resize_bilinear_grad(
      grad,
      op.inputs[0],
      align_corners=op.get_attr("align_corners"),
      half_pixel_centers=op.get_attr("half_pixel_centers"))
  return [grad0, None]


@ops.RegisterGradient("ScaleAndTranslate")
def _ScaleAndTranslateGrad(op, grad):
  """The derivatives for ScaleAndTranslate transformation op.

  Args:
    op: The ScaleAndTranslate op.
    grad: The tensor representing the gradient w.r.t. the output.

  Returns:
    The gradients w.r.t. the input.
  """

  grad0 = gen_image_ops.scale_and_translate_grad(
      grad,
      op.inputs[0],
      op.inputs[2],
      op.inputs[3],
      kernel_type=op.get_attr("kernel_type"),
      antialias=op.get_attr("antialias"))
  return [grad0, None, None, None]


@ops.RegisterGradient("ResizeBicubic")
def _ResizeBicubicGrad(op, grad):
  """The derivatives for bicubic resizing.

  Args:
    op: The ResizeBicubic op.
    grad: The tensor representing the gradient w.r.t. the output.

  Returns:
    The gradients w.r.t. the input.
  """
  allowed_types = [dtypes.float32, dtypes.float64]
  grad0 = None
  if op.inputs[0].dtype in allowed_types:
    grad0 = gen_image_ops.resize_bicubic_grad(
        grad,
        op.inputs[0],
        align_corners=op.get_attr("align_corners"),
        half_pixel_centers=op.get_attr("half_pixel_centers"))
  return [grad0, None]


@ops.RegisterGradient("CropAndResize")
def _CropAndResizeGrad(op, grad):
  """The derivatives for crop_and_resize.

  We back-propagate to the image only when the input image tensor has floating
  point dtype but we always back-propagate to the input boxes tensor.

  Args:
    op: The CropAndResize op.
    grad: The tensor representing the gradient w.r.t. the output.

  Returns:
    The gradients w.r.t. the input image, boxes, as well as the always-None
    gradients w.r.t. box_ind and crop_size.
  """
  image = op.inputs[0]
  if image.get_shape().is_fully_defined():
    image_shape = image.get_shape().as_list()
  else:
    image_shape = array_ops.shape(image)

  allowed_types = [dtypes.float16, dtypes.float32, dtypes.float64]
  if op.inputs[0].dtype in allowed_types:
    # pylint: disable=protected-access
    grad0 = gen_image_ops.crop_and_resize_grad_image(
        grad, op.inputs[1], op.inputs[2], image_shape, T=op.get_attr("T"),
        method=op.get_attr("method"))
    # pylint: enable=protected-access
  else:
    grad0 = None

  # `grad0` is the gradient to the input image pixels and it
  # has been implemented for nearest neighbor and bilinear sampling
  # respectively. `grad1` is the gradient to the input crop boxes' coordinates.
  # When using nearest neighbor sampling, the gradient to crop boxes'
  # coordinates are not well defined. In practice, we still approximate
  # grad1 using the gradient derived from bilinear sampling.
  grad1 = gen_image_ops.crop_and_resize_grad_boxes(
      grad, op.inputs[0], op.inputs[1], op.inputs[2])

  return [grad0, grad1, None, None]


def _CustomReciprocal(x):
  """Wrapper function around `math_ops.div_no_nan()` to perform a "safe" reciprocal incase the input is zero. Avoids divide by zero and NaNs.

  Input:
    x -> input tensor to be reciprocat-ed.
  Returns:
    x_reciprocal -> reciprocal of x without NaNs.
  """
  return math_ops.div_no_nan(1.0, x)


@ops.RegisterGradient("RGBToHSV")
def _RGBToHSVGrad(op, grad):
  """The gradients for `rgb_to_hsv` operation.

  This function is a piecewise continuous function as defined here:
  https://en.wikipedia.org/wiki/HSL_and_HSV#From_RGB
  We perform the multi variate derivative and compute all partial derivates
  seperately before adding them in the end. Formulas are given before each
  partial derivative calculation.

  Args:
    op: The `rgb_to_hsv` `Operation` that we are differentiating.
    grad: Gradient with respect to the output of the `rgb_to_hsv` op.

  Returns:
    Gradients with respect to the input of `rgb_to_hsv`.
  """
  # Input Channels
  reds = op.inputs[0][..., 0]
  greens = op.inputs[0][..., 1]
  blues = op.inputs[0][..., 2]
  # Output Channels
  saturation = op.outputs[0][..., 1]
  value = op.outputs[0][..., 2]

  # Mask/Indicator for max and min values of each pixel.
  # Arbitrary assignment in case of tie breakers with R>G>B.
  # Max values
  red_biggest = math_ops.cast((reds >= blues) & \
                 (reds >= greens), dtypes.float32)
  green_biggest = math_ops.cast((greens > reds) & \
                   (greens >= blues), dtypes.float32)
  blue_biggest = math_ops.cast((blues > reds) & \
                  (blues > greens), dtypes.float32)
  # Min values
  red_smallest = math_ops.cast((reds < blues) & \
                  (reds < greens), dtypes.float32)
  green_smallest = math_ops.cast((greens <= reds) & \
                    (greens < blues), dtypes.float32)
  blue_smallest = math_ops.cast((blues <= reds) & \
                   (blues <= greens), dtypes.float32)

  # Derivatives of R, G, B wrt Value slice
  dv_dr = red_biggest
  dv_dg = green_biggest
  dv_db = blue_biggest

  # Derivatives of R, G, B wrt Saturation slice

  # The first term in the addition is the case when the corresponding color
  # from (r,g,b) was "MAX"
  # -> derivative = MIN/square(MAX), MIN could be one of the other two colors
  # The second term is the case when the corresponding color from
  # (r,g,b) was "MIN"
  # -> derivative = -1/MAX, MAX could be one of the other two colours.
  ds_dr = math_ops.cast(reds > 0, dtypes.float32) * \
               math_ops.add(red_biggest * \
               math_ops.add(green_smallest * greens, blue_smallest * blues) * \
               _CustomReciprocal(math_ops.square(reds)),\
               red_smallest * -1 * _CustomReciprocal((green_biggest * \
               greens) + (blue_biggest * blues)))
  ds_dg = math_ops.cast(greens > 0, dtypes.float32) * \
               math_ops.add(green_biggest * \
               math_ops.add(red_smallest * reds, blue_smallest * blues) * \
               _CustomReciprocal(math_ops.square(greens)),\
               green_smallest * -1 * _CustomReciprocal((red_biggest * \
               reds) + (blue_biggest * blues)))
  ds_db = math_ops.cast(blues > 0, dtypes.float32) * \
               math_ops.add(blue_biggest * \
               math_ops.add(green_smallest * greens, red_smallest * reds) * \
               _CustomReciprocal(math_ops.square(blues)),\
               blue_smallest * -1 * _CustomReciprocal((green_biggest * \
               greens) + (red_biggest * reds)))

  # Derivatives of R, G, B wrt Hue slice

  # Need to go case by case for each color.
  # for red, dh_dr -> dh_dr_1 + dh_dr_2 + dh_dr_3 + dh_dr_4 + dh_dr_5
  # dh_dr_1 ->
  # if red was MAX, then derivative = 60 * -1 * (G-B)/square(MAX-MIN) == 60 *\
  #  -1 * (greens-blues) * reciprocal(square(saturation)) * \
  #  reciprical(square(value))
  # elif green was MAX, there are two subcases
  # ie when red was MIN and when red was NOT MIN
  #   dh_dr_2 ->
  #   if red was MIN (use UV rule) ->  60 * ((1 * -1/(MAX-MIN)) +\
  #     (B-R)*(-1/square(MAX-MIN) * -1)) == 60 * (blues - greens) *\
  #     reciprocal(square(reds - greens))
  #   dh_dr_3 ->
  #   if red was NOT MIN -> 60 * -1/MAX-MIN == -60 * reciprocal(greens-blues)
  # elif blue was MAX, there are two subcases
  #   dh_dr_4 ->
  #   if red was MIN (similarly use the UV rule) -> 60 * (blues - greens) *\
  #     reciprocal(square(blues - reds))
  #   dh_dr_5 ->
  #   if red was NOT MIN -> 60 * 1/MAX-MIN == 60 * reciprocal(blues-greens)
  dh_dr_1 = 60 * (math_ops.cast(reds > 0, dtypes.float32) * red_biggest * \
                  -1  * \
                  (greens - blues) * \
                  _CustomReciprocal(math_ops.square(saturation)) *\
                  _CustomReciprocal(math_ops.square(value)))
  dh_dr_2 = 60 * (math_ops.cast(greens > 0, dtypes.float32) * green_biggest * \
                  red_smallest  * (blues - greens) * \
                  _CustomReciprocal(math_ops.square(reds - greens)))
  dh_dr_3 = 60 * (math_ops.cast(greens > 0, dtypes.float32) * green_biggest * \
                  blue_smallest * -1 * _CustomReciprocal(greens - blues))
  dh_dr_4 = 60 * (math_ops.cast(blues > 0, dtypes.float32) * blue_biggest * \
                  red_smallest * (blues - greens) * \
                  _CustomReciprocal(math_ops.square(blues - reds)))
  dh_dr_5 = 60 * (math_ops.cast(blues > 0, dtypes.float32) * blue_biggest * \
                  green_smallest * _CustomReciprocal(blues - greens))

  dh_dr = dh_dr_1 + dh_dr_2 + dh_dr_3 + dh_dr_4 + dh_dr_5
  # Converting from degrees to [0,1] scale as specified in
  # https://www.tensorflow.org/api_docs/python/tf/image/rgb_to_hsv
  dh_dr = dh_dr / 360

  # for green, dh_dg -> dh_dg_1 + dh_dg_2 + dh_dg_3 + dh_dg_4 + dh_dg_5
  # dh_dg_1 ->
  # if green was MAX, then derivative = 60 * -1 * (B-R)/square(MAX-MIN) == 60 *\
  #   -1 * (blues - reds) * reciprocal(square(saturation)) * \
  #   reciprocal(square(value))
  # elif red was MAX, there are two subcases ie
  # when green was MIN and when green was NOT MIN
  #   dh_dg_2 ->
  #   if green was MIN (use UV rule) ->  60 * ((1 * 1/(MAX-MIN)) + \
  #     (greens-blues) * (-1/square(MAX-MIN) * -1)) == 60 * \
  #     ((reciprocal(reds-greens) + (greens-blues) * \
  #     reciprocal(square(reds-greens))))
  #   dh_dg_3 ->
  #   if green was NOT MIN -> 60 * 1/MAX-MIN == 60 * reciprocal(reds - blues)
  # elif blue was MAX, there are two subcases
  #   dh_dg_4 ->
  #   if green was MIN (similarly use the UV rule) -> 60 * -1 * \
  #     (reciprocal(blues - greens) + (reds-greens)* -1 * \
  #     reciprocal(square(blues-greens)))
  #   dh_dr_5 ->
  #   if green was NOT MIN -> 60 * -1/MAX-MIN == -60 * reciprocal(blues - reds)
  dh_dg_1 = 60 * (math_ops.cast(greens > 0, dtypes.float32) * green_biggest * \
                  -1 * (blues - reds) * \
                  _CustomReciprocal(math_ops.square(saturation))\
                  * _CustomReciprocal(math_ops.square(value)))
  dh_dg_2 = 60 * (math_ops.cast(reds > 0, dtypes.float32) * red_biggest * \
                  green_smallest * (reds - blues) * \
                  _CustomReciprocal(math_ops.square(reds - greens)))
  dh_dg_3 = 60 * (math_ops.cast(reds > 0, dtypes.float32) * red_biggest * \
                  blue_smallest * _CustomReciprocal(reds - blues))
  dh_dg_4 = 60 * (math_ops.cast(blues > 0, dtypes.float32) * blue_biggest * \
                  green_smallest * (reds - blues) * \
                  _CustomReciprocal(math_ops.square(blues - greens)))
  dh_dg_5 = 60 * (math_ops.cast(blues > 0, dtypes.float32) * blue_biggest * \
                  red_smallest * -1 * _CustomReciprocal(blues - reds))

  dh_dg = dh_dg_1 + dh_dg_2 + dh_dg_3 + dh_dg_4 + dh_dg_5
  # Converting from degrees to [0,1] scale as specified in
  # https://www.tensorflow.org/api_docs/python/tf/image/rgb_to_hsv
  dh_dg = dh_dg / 360

  # for blue, dh_db -> dh_db_1 + dh_db_2 + dh_db_3 + dh_db_4 + dh_db_5
  # dh_db_1 ->
  # if blue was MAX, then derivative = 60 * -1 * (R-G)/square(MAX-MIN) == 60 *\
  #   -1 * reciprocal(square(saturation)) * reciprocal(square(value))
  # elif red was MAX, there are two subcases
  # ie when blue was MIN and when blue was NOT MIN
  #   dh_dg_2 ->
  #   if blue was MIN (use UV rule) ->  60 * ((1 * -1/(MAX-MIN)) + \
  #     (greens-blues) * (-1/square(MAX-MIN) * -1)) == 60 * (greens - reds) *\
  #     reciprocal(square(reds - blues))
  #   dh_dg_3 ->
  #   if blue was NOT MIN -> 60 * -1/MAX-MIN == 60 * -1 * \
  #     reciprocal(reds - greens)
  # elif green was MAX, there are two subcases
  #   dh_dg_4 ->
  #   if blue was MIN (similarly use the UV rule) -> 60 * -1 * \
  #     (reciprocal(greens - blues) + (blues - reds) * -1 * \
  #     reciprocal(square(greens - blues)))
  #   dh_dr_5 ->
  #   if blue was NOT MIN -> 60 * 1/MAX-MIN == 60 * reciprocal(greens - reds)
  dh_db_1 = 60 * (math_ops.cast(blues > 0, dtypes.float32) * blue_biggest * \
                   -1 * \
                  (reds - greens) * \
                  _CustomReciprocal(math_ops.square(saturation)) * \
                  _CustomReciprocal(math_ops.square(value)))
  dh_db_2 = 60 * (math_ops.cast(reds > 0, dtypes.float32) * red_biggest *\
                  blue_smallest * (greens - reds) * \
                  _CustomReciprocal(math_ops.square(reds - blues)))
  dh_db_3 = 60 * (math_ops.cast(reds > 0, dtypes.float32) * red_biggest * \
                  green_smallest * -1 * _CustomReciprocal(reds - greens))
  dh_db_4 = 60 * (math_ops.cast(greens > 0, dtypes.float32) * green_biggest * \
                  blue_smallest * (greens - reds) * \
                  _CustomReciprocal(math_ops.square(greens - blues)))
  dh_db_5 = 60 * (math_ops.cast(greens > 0, dtypes.float32) * green_biggest * \
                  red_smallest * _CustomReciprocal(greens - reds))

  dh_db = dh_db_1 + dh_db_2 + dh_db_3 + dh_db_4 + dh_db_5
  # Converting from degrees to [0,1] scale as specified in
  # https://www.tensorflow.org/api_docs/python/tf/image/rgb_to_hsv
  dh_db = dh_db / 360

  # Gradients wrt to inputs
  dv_drgb = array_ops.stack(
      [grad[..., 2] * dv_dr, grad[..., 2] * dv_dg, grad[..., 2] * dv_db],
      axis=-1)
  ds_drgb = array_ops.stack(
      [grad[..., 1] * ds_dr, grad[..., 1] * ds_dg, grad[..., 1] * ds_db],
      axis=-1)
  dh_drgb = array_ops.stack(
      [grad[..., 0] * dh_dr, grad[..., 0] * dh_dg, grad[..., 0] * dh_db],
      axis=-1)

  gradient_input = math_ops.add(math_ops.add(dv_drgb, ds_drgb), dh_drgb)
  return gradient_input



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Ignite File System for checkpointing and communication with TensorBoard.

Apache Ignite is a memory-centric distributed database, caching, and
processing platform for transactional, analytical, and streaming workloads,
delivering in-memory speeds at petabyte scale. In addition to database
functionality Apache Ignite provides a distributed file system called
IGFS (https://ignite.apache.org/features/igfs.html). IGFS delivers a similar
functionality to Hadoop HDFS, but only in-memory. In fact, in addition to
its own APIs, IGFS implements Hadoop FileSystem API and can be transparently
plugged into Hadoop or Spark deployments. This contrib package contains an
integration between IGFS and TensorFlow.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from tensorflow.contrib.ignite.python.ops import ignite_op_loader  # pylint: disable=unused-import
from tensorflow.python.framework import load_library
from tensorflow.python.platform import resource_loader

file_system_library = os.path.join(resource_loader.get_data_files_path(),
                                   "../../_ignite_ops.so")
load_library.load_file_system_library(file_system_library)



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Experimental `dataset` API for parsing example."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.data.util import structure
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.framework import sparse_tensor
from tensorflow.python.framework import tensor_spec
from tensorflow.python.ops import gen_experimental_dataset_ops
from tensorflow.python.ops import parsing_ops
from tensorflow.python.util.tf_export import tf_export


class _ParseExampleDataset(dataset_ops.UnaryDataset):
  """A `Dataset` that parses `example` dataset into a `dict` dataset."""

  def __init__(self, input_dataset, features, num_parallel_calls):
    self._input_dataset = input_dataset
    if not structure.are_compatible(
        input_dataset.element_spec,
        tensor_spec.TensorSpec([None], dtypes.string)):
      raise TypeError("Input dataset should be a dataset of vectors of strings")
    self._num_parallel_calls = num_parallel_calls
    # pylint: disable=protected-access
    self._features = parsing_ops._prepend_none_dimension(features)
    # sparse_keys and dense_keys come back sorted here.
    (sparse_keys, sparse_types, dense_keys, dense_types, dense_defaults,
     dense_shapes) = parsing_ops._features_to_raw_params(
         self._features, [
             parsing_ops.VarLenFeature, parsing_ops.SparseFeature,
             parsing_ops.FixedLenFeature, parsing_ops.FixedLenSequenceFeature
         ])
    # TODO(b/112859642): Pass sparse_index and sparse_values for SparseFeature.
    (_, dense_defaults_vec, sparse_keys, sparse_types, dense_keys, dense_shapes,
     dense_shape_as_shape) = parsing_ops._process_raw_parameters(
         None, dense_defaults, sparse_keys, sparse_types, dense_keys,
         dense_types, dense_shapes)
    # pylint: enable=protected-access
    self._sparse_keys = sparse_keys
    self._sparse_types = sparse_types
    self._dense_keys = dense_keys
    self._dense_defaults = dense_defaults_vec
    self._dense_shapes = dense_shapes
    self._dense_types = dense_types
    input_dataset_shape = dataset_ops.get_legacy_output_shapes(
        self._input_dataset)
    dense_output_shapes = [input_dataset_shape.concatenate(shape)
                           for shape in dense_shape_as_shape]
    sparse_output_shapes = [input_dataset_shape.concatenate([None])
                            for _ in range(len(sparse_keys))]

    output_shapes = dict(
        zip(self._dense_keys + self._sparse_keys,
            dense_output_shapes + sparse_output_shapes))
    output_types = dict(
        zip(self._dense_keys + self._sparse_keys,
            self._dense_types + self._sparse_types))
    output_classes = dict(
        zip(self._dense_keys + self._sparse_keys,
            [ops.Tensor for _ in range(len(self._dense_defaults))] +
            [sparse_tensor.SparseTensor for _ in range(len(self._sparse_keys))
            ]))
    self._element_spec = structure.convert_legacy_structure(
        output_types, output_shapes, output_classes)

    variant_tensor = (
        gen_experimental_dataset_ops.parse_example_dataset(
            self._input_dataset._variant_tensor,  # pylint: disable=protected-access
            self._num_parallel_calls,
            self._dense_defaults,
            self._sparse_keys,
            self._dense_keys,
            self._sparse_types,
            self._dense_shapes,
            **self._flat_structure))
    super(_ParseExampleDataset, self).__init__(input_dataset, variant_tensor)

  @property
  def element_spec(self):
    return self._element_spec


# TODO(b/111553342): add arguments names and example names as well.
@tf_export("data.experimental.parse_example_dataset")
def parse_example_dataset(features, num_parallel_calls=1):
  """A transformation that parses `Example` protos into a `dict` of tensors.

  Parses a number of serialized `Example` protos given in `serialized`. We refer
  to `serialized` as a batch with `batch_size` many entries of individual
  `Example` protos.

  This op parses serialized examples into a dictionary mapping keys to `Tensor`
  and `SparseTensor` objects. `features` is a dict from keys to `VarLenFeature`,
  `SparseFeature`, and `FixedLenFeature` objects. Each `VarLenFeature`
  and `SparseFeature` is mapped to a `SparseTensor`, and each
  `FixedLenFeature` is mapped to a `Tensor`. See `tf.io.parse_example` for more
  details about feature dictionaries.

  Args:
   features: A `dict` mapping feature keys to `FixedLenFeature`,
     `VarLenFeature`, and `SparseFeature` values.
   num_parallel_calls: (Optional.) A `tf.int32` scalar `tf.Tensor`,
      representing the number of parsing processes to call in parallel.

  Returns:
    A dataset transformation function, which can be passed to
    `tf.data.Dataset.apply`.

  Raises:
    ValueError: if features argument is None.
  """
  if features is None:
    raise ValueError("Missing: features was %s." % features)

  def _apply_fn(dataset):
    """Function from `Dataset` to `Dataset` that applies the transformation."""
    out_dataset = _ParseExampleDataset(dataset, features, num_parallel_calls)
    if any(
        isinstance(feature, parsing_ops.SparseFeature)
        for _, feature in features.items()
    ):
      # pylint: disable=protected-access
      # pylint: disable=g-long-lambda
      out_dataset = out_dataset.map(
          lambda x: parsing_ops._construct_sparse_tensors_for_sparse_features(
              features, x), num_parallel_calls=num_parallel_calls)
    return out_dataset

  return _apply_fn



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Training related logic for Keras model in TF 2.0 context.

Note that all the code under this module is under active development, please DO
NOT use it unless you are really sure what you are doing.
"""

# pylint: disable=protected-access
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools

from tensorflow.python.distribute import distribution_strategy_context
from tensorflow.python.eager import def_function
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_util
from tensorflow.python.framework.ops import composite_tensor
from tensorflow.python.keras import backend
from tensorflow.python.keras.distribute import distributed_training_utils as dist_utils
from tensorflow.python.keras.engine import training_eager
from tensorflow.python.keras.engine import training_utils
from tensorflow.python.keras.utils.mode_keys import ModeKeys
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.util import nest


def _get_or_make_execution_function(model, mode):
  """Makes or reuses function to run one step of distributed model execution."""
  model._init_distributed_function_cache_if_not_compiled()

  # Use a key with 'v2' to distinguish from fall-back execution functions.
  key = (mode, 'v2')
  distributed_function = dist_utils.get_distributed_function(model, key)
  if distributed_function:
    return distributed_function

  distribution_function = _make_execution_function(model, mode)
  dist_utils.set_distributed_function(model, key, distribution_function)
  return distribution_function


def _make_execution_function(model, mode):
  """Creates a function to run one step of distributed model execution."""
  per_replica_function = _make_replica_execution_function(model, mode)

  def distributed_function(input_iterator):
    """A single step of the distributed execution across replicas."""
    x, y, sample_weights = _prepare_feed_values(
        model, input_iterator, mode)
    # Call `Model.{train,test,predict}_on_batch` on every replica passing
    # PerReplicas as arguments.  On every replica inside this call, each
    # PerReplica object will return the value for that replica.  The outputs
    # are PerReplicas too.
    strategy = distribution_strategy_context.get_strategy()
    outputs = strategy.experimental_run_v2(
        per_replica_function, args=(x, y, sample_weights))
    # Out of PerReplica outputs reduce or pick values to return.
    all_outputs = dist_utils.unwrap_output_dict(
        strategy, outputs, mode)
    return all_outputs

  if not model.run_eagerly:
    distributed_function = def_function.function(
        distributed_function, autograph=False)

  def execution_function(input_fn):
    # `numpy` translates Tensors to values in Eager mode.
    return nest.map_structure(_non_none_constant_value,
                              distributed_function(input_fn))

  return execution_function


def _non_none_constant_value(v):
  constant_value = tensor_util.constant_value(v)
  return constant_value if constant_value is not None else v


def _prepare_feed_values(model, inputs, mode):
  """Prepare feed values to the model execution function.

  Arguments:
    model: Model to prepare feed values for.
    inputs: An iterator of model inputs, targets, and sample_weights.
      model inputs may be lists, single values, or dicts mapping input feed
      names to values.
    mode: One of ModeKeys.TRAIN/ModeKeys.TEST/ModeKeys.PREDICT.

  Returns:
    Feed values for the model in the given mode. This is a tuple of
    the structure (inputs, targets, sample_weights), where each of
    (tuple, targets, sample_weights) may be a python list. Single values
    for inputs will always be wrapped in lists.
  """
  inputs, targets, sample_weights = _get_input_from_iterator(inputs)

  # When the inputs are dict, then we want to flatten it in the same order as
  # the input layers, such that the data are fed into the input layers in the
  # correct order.
  if isinstance(inputs, dict):
    inputs = [inputs[key] for key in model._feed_input_names]
  else:
    inputs = training_utils.ModelInputs(inputs).as_list()

  if mode == ModeKeys.PREDICT:
    sample_weights = []
    targets = []

  ins = [inputs, targets, sample_weights]
  return tuple(ins)


def _get_input_from_iterator(iterator):
  """Get elements from the iterator and verify the input shape and type."""
  next_element = next(iterator)

  if (tensor_util.is_tensor(next_element) or
      isinstance(next_element, (dict, composite_tensor.CompositeTensor))):
    next_element = [next_element]
  if len(next_element) == 1:
    x, = next_element
    y = None
    sample_weights = None
  elif len(next_element) == 2:
    x, y = next_element
    sample_weights = None
  else:
    x, y, sample_weights = next_element

  # Validate that all the elements in x and y are of the same type and shape.
  dist_utils.validate_distributed_dataset_inputs(
      distribution_strategy_context.get_strategy(), x, y, sample_weights)
  return x, y, sample_weights


def _make_replica_execution_function(model, mode):
  """A single step of the distributed execution on a replica."""
  if mode == ModeKeys.TRAIN:
    func = functools.partial(train_on_batch, model)
  elif mode == ModeKeys.TEST:
    func = functools.partial(test_on_batch, model)
  else:
    def _predict_on_batch(x, y=None, sample_weights=None):
      del y, sample_weights
      return predict_on_batch(model, x)

    func = _predict_on_batch

  if mode != ModeKeys.PREDICT:
    # `reset_metrics` is set to False to maintain stateful metrics across
    # batch-level calls.
    func = functools.partial(func, reset_metrics=False)

  return func


def _prepare_model_with_inputs(model, dataset):
  """Use the data from the adapter to config the model.

  Model need to be properly configured before training, eg build with inputs, or
  compile with inputs for subclass model.

  Args:
    model: a Keras model object.
    dataset: a eager dataset instance where the data will be extracted.
  """
  if not model.inputs:
    inputs, target, _ = model._build_model_with_inputs(dataset, targets=None)
  else:
    inputs, target, _ = _get_input_from_iterator(iter(dataset))

  if not model._is_compiled and model.optimizer:
    model._compile_from_inputs(inputs, target, dataset, None)

  if target is not None:
    training_utils.prepare_sample_weight_modes(model._training_endpoints,
                                               model.sample_weight_mode)


def train_on_batch(
    model,
    x,
    y=None,
    sample_weight=None,
    class_weight=None,
    reset_metrics=True):
  """Runs a single gradient update on a single batch of data.

  Arguments:
      model: The model to train.
      x: Input data. It could be:
        - A Numpy array (or array-like), or a list of arrays
            (in case the model has multiple inputs).
        - A TensorFlow tensor, or a list of tensors
            (in case the model has multiple inputs).
        - A dict mapping input names to the corresponding array/tensors,
            if the model has named inputs.
        - A `tf.data` dataset.
      y: Target data. Like the input data `x`, it could be either Numpy
        array(s) or TensorFlow tensor(s). It should be consistent with `x`
        (you cannot have Numpy inputs and tensor targets, or inversely). If
        `x` is a dataset `y` should not be specified
        (since targets will be obtained from the iterator).
      sample_weight: Optional array of the same length as x, containing
        weights to apply to the model's loss for each sample. In the case of
        temporal data, you can pass a 2D array with shape (samples,
        sequence_length), to apply a different weight to every timestep of
        every sample. In this case you should make sure to specify
        sample_weight_mode="temporal" in compile(). This argument is not
        supported when `x` is a dataset.
      class_weight: Optional dictionary mapping class indices (integers) to a
        weight (float) to apply to the model's loss for the samples from this
        class during training. This can be useful to tell the model to "pay
        more attention" to samples from an under-represented class.
      reset_metrics: If `True`, the metrics returned will be only for this
        batch. If `False`, the metrics will be statefully accumulated across
        batches.

  Returns:
      Scalar training loss
      (if the model has a single output and no metrics)
      or list of scalars (if the model has multiple outputs
      and/or metrics). The attribute `model.metrics_names` will give you
      the display labels for the scalar outputs.

  Raises:
    ValueError: In case of invalid user-provided arguments.
  """
  model._assert_compile_was_called()

  # TODO(scottzhu): Standardization should happen in the data handlers,
  ## not on a per batch basis in the *_on_batch methods
  # Validate and standardize user data.
  x, y, sample_weights = model._standardize_user_data(
      x, y, sample_weight=sample_weight, class_weight=class_weight,
      extract_tensors_from_dataset=True)
  batch_size = array_ops.shape(nest.flatten(x, expand_composites=True)[0])[0]
  # If `model._distribution_strategy` is True, then we are in a replica context
  # at this point because of the check above.  `train_on_batch` is being run
  # for each replica by `model._distribution_strategy` and the same code path
  # as Eager is expected to be taken.
  outputs = training_eager.train_on_batch(
      model,
      x,
      y,
      sample_weights=sample_weights,
      output_loss_metrics=model._output_loss_metrics)

  if reset_metrics:
    model.reset_metrics()

  outputs['batch_size'] = math_ops.cast(batch_size, dtypes.int64)
  return outputs


def test_on_batch(model, x, y=None, sample_weight=None, reset_metrics=True):
  """Test the model on a single batch of samples.

  Arguments:
      model: The model to test.
      x: Input data. It could be:
        - A Numpy array (or array-like), or a list of arrays
          (in case the model has multiple inputs).
        - A TensorFlow tensor, or a list of tensors
          (in case the model has multiple inputs).
        - A dict mapping input names to the corresponding array/tensors,
          if the model has named inputs.
        - A `tf.data` dataset.
      y: Target data. Like the input data `x`,
        it could be either Numpy array(s) or TensorFlow tensor(s).
        It should be consistent with `x` (you cannot have Numpy inputs and
        tensor targets, or inversely). If `x` is a dataset,
        `y` should not be specified
        (since targets will be obtained from the iterator).
      sample_weight: Optional array of the same length as x, containing
          weights to apply to the model's loss for each sample.
          In the case of temporal data, you can pass a 2D array
          with shape (samples, sequence_length),
          to apply a different weight to every timestep of every sample.
          In this case you should make sure to specify
          sample_weight_mode="temporal" in compile(). This argument is not
          supported when `x` is a dataset.
      reset_metrics: If `True`, the metrics returned will be only for this
        batch. If `False`, the metrics will be statefully accumulated across
        batches.

  Returns:
      Scalar test loss (if the model has a single output and no metrics)
      or list of scalars (if the model has multiple outputs
      and/or metrics). The attribute `model.metrics_names` will give you
      the display labels for the scalar outputs.

  Raises:
      ValueError: In case of invalid user-provided arguments.
  """
  model._assert_compile_was_called()

  # TODO(scottzhu): Standardization should happen in the data handlers,
  ## not on a per batch basis in the *_on_batch methods
  # Validate and standardize user data.
  x, y, sample_weights = model._standardize_user_data(
      x, y, sample_weight=sample_weight, extract_tensors_from_dataset=True)

  batch_size = array_ops.shape(nest.flatten(x, expand_composites=True)[0])[0]
  outputs = training_eager.test_on_batch(
      model,
      x,
      y,
      sample_weights=sample_weights,
      output_loss_metrics=model._output_loss_metrics)

  if reset_metrics:
    model.reset_metrics()

  outputs['batch_size'] = math_ops.cast(batch_size, dtypes.int64)
  return outputs


def predict_on_batch(model, x):
  """Returns predictions for a single batch of samples.

  Arguments:
      model: The model to predict with.
      x: Input data. It could be:
        - A Numpy array (or array-like), or a list of arrays
          (in case the model has multiple inputs).
        - A TensorFlow tensor, or a list of tensors
          (in case the model has multiple inputs).
        - A `tf.data` dataset.

  Returns:
      Numpy array(s) of predictions.

  Raises:
      ValueError: In case of mismatch between given number of inputs and
        expectations of the model.
  """
  # TODO(scottzhu): Standardization should happen in the data handlers,
  ## not on a per batch basis in the *_on_batch methods
  # Validate and standardize user data.
  inputs, _, _ = model._standardize_user_data(
      x, extract_tensors_from_dataset=True)

  # If `model._distribution_strategy` is True, then we are in a replica context
  # at this point.
  inputs = training_utils.cast_if_floating_dtype(inputs)
  if isinstance(inputs, collections.Sequence):
    # Unwrap lists with only one input, as we do when training on batch
    if len(inputs) == 1:
      inputs = inputs[0]

  with backend.eager_learning_phase_scope(0):
    return model(inputs)  # pylint: disable=not-callable



"""Python wrappers around TensorFlow ops.

This file is MACHINE GENERATED! Do not edit.
Original C++ source file: audio_microfrontend_op.cc
"""

import collections as _collections
import six as _six

from tensorflow.python import pywrap_tensorflow as _pywrap_tensorflow
from tensorflow.python.eager import context as _context
from tensorflow.python.eager import core as _core
from tensorflow.python.eager import execute as _execute
from tensorflow.python.framework import dtypes as _dtypes
from tensorflow.python.framework import errors as _errors
from tensorflow.python.framework import tensor_shape as _tensor_shape

from tensorflow.core.framework import op_def_pb2 as _op_def_pb2
# Needed to trigger the call to _set_call_cpp_shape_fn.
from tensorflow.python.framework import common_shapes as _common_shapes
from tensorflow.python.framework import op_def_registry as _op_def_registry
from tensorflow.python.framework import ops as _ops
from tensorflow.python.framework import op_def_library as _op_def_library
from tensorflow.python.util.deprecation import deprecated_endpoints
from tensorflow.python.util import dispatch as _dispatch
from tensorflow.python.util.tf_export import tf_export
from tensorflow.python.util.tf_export import kwarg_only as _kwarg_only
from tensorflow.tools.docs import doc_controls as _doc_controls


@_dispatch.add_dispatch_list
@tf_export('audio_microfrontend')
def audio_microfrontend(audio, sample_rate=16000, window_size=25, window_step=10, num_channels=32, upper_band_limit=7500, lower_band_limit=125, smoothing_bits=10, even_smoothing=0.025, odd_smoothing=0.06, min_signal_remaining=0.05, enable_pcan=False, pcan_strength=0.95, pcan_offset=80, gain_bits=21, enable_log=True, scale_shift=6, left_context=0, right_context=0, frame_stride=1, zero_padding=False, out_scale=1, out_type=_dtypes.uint16, name=None):
  r"""Audio Microfrontend Op.

  This Op converts a sequence of audio data into one or more
  feature vectors containing filterbanks of the input. The
  conversion process uses a lightweight library to perform:

  1. A slicing window function
  2. Short-time FFTs
  3. Filterbank calculations
  4. Noise reduction
  5. PCAN Auto Gain Control
  6. Logarithmic scaling

  Arguments
    audio: 1D Tensor, int16 audio data in temporal ordering.
    sample_rate: Integer, the sample rate of the audio in Hz.
    window_size: Integer, length of desired time frames in ms.
    window_step: Integer, length of step size for the next frame in ms.
    num_channels: Integer, the number of filterbank channels to use.
    upper_band_limit: Float, the highest frequency included in the filterbanks.
    lower_band_limit: Float, the lowest frequency included in the filterbanks.
    smoothing_bits: Int, scale up signal by 2^(smoothing_bits) before reduction.
    even_smoothing: Float, smoothing coefficient for even-numbered channels.
    odd_smoothing: Float, smoothing coefficient for odd-numbered channels.
    min_signal_remaining: Float, fraction of signal to preserve in smoothing.
    enable_pcan: Bool, enable PCAN auto gain control.
    pcan_strength: Float, gain normalization exponent.
    pcan_offset: Float, positive value added in the normalization denominator.
    gain_bits: Int, number of fractional bits in the gain.
    enable_log: Bool, enable logarithmic scaling of filterbanks.
    scale_shift: Integer, scale filterbanks by 2^(scale_shift).
    left_context: Integer, number of preceding frames to attach to each frame.
    right_context: Integer, number of preceding frames to attach to each frame.
    frame_stride: Integer, M frames to skip over, where output[n] = frame[n*M].
    zero_padding: Bool, if left/right context is out-of-bounds, attach frame of
                  zeroes. Otherwise, frame[0] or frame[size-1] will be copied.
    out_scale: Integer, divide all filterbanks by this number.
    out_type: DType, type of the output Tensor, defaults to UINT16.

  Returns
    filterbanks: 2D Tensor, each row is a time frame, each column is a channel.

  Args:
    audio: A `Tensor` of type `int16`.
    sample_rate: An optional `int`. Defaults to `16000`.
    window_size: An optional `int`. Defaults to `25`.
    window_step: An optional `int`. Defaults to `10`.
    num_channels: An optional `int`. Defaults to `32`.
    upper_band_limit: An optional `float`. Defaults to `7500`.
    lower_band_limit: An optional `float`. Defaults to `125`.
    smoothing_bits: An optional `int`. Defaults to `10`.
    even_smoothing: An optional `float`. Defaults to `0.025`.
    odd_smoothing: An optional `float`. Defaults to `0.06`.
    min_signal_remaining: An optional `float`. Defaults to `0.05`.
    enable_pcan: An optional `bool`. Defaults to `False`.
    pcan_strength: An optional `float`. Defaults to `0.95`.
    pcan_offset: An optional `float`. Defaults to `80`.
    gain_bits: An optional `int`. Defaults to `21`.
    enable_log: An optional `bool`. Defaults to `True`.
    scale_shift: An optional `int`. Defaults to `6`.
    left_context: An optional `int`. Defaults to `0`.
    right_context: An optional `int`. Defaults to `0`.
    frame_stride: An optional `int`. Defaults to `1`.
    zero_padding: An optional `bool`. Defaults to `False`.
    out_scale: An optional `int`. Defaults to `1`.
    out_type: An optional `tf.DType` from: `tf.uint16, tf.float32`. Defaults to `tf.uint16`.
    name: A name for the operation (optional).

  Returns:
    A `Tensor` of type `out_type`.
  """
  _ctx = _context._context or _context.context()
  if _ctx is not None and _ctx._thread_local_data.is_eager:
    try:
      _result = _pywrap_tensorflow.TFE_Py_FastPathExecute(
        _ctx._context_handle, _ctx._thread_local_data.device_name,
        "AudioMicrofrontend", name, _ctx.post_execution_callbacks, audio,
        "sample_rate", sample_rate, "window_size", window_size, "window_step",
        window_step, "num_channels", num_channels, "upper_band_limit",
        upper_band_limit, "lower_band_limit", lower_band_limit,
        "smoothing_bits", smoothing_bits, "even_smoothing", even_smoothing,
        "odd_smoothing", odd_smoothing, "min_signal_remaining",
        min_signal_remaining, "enable_pcan", enable_pcan, "pcan_strength",
        pcan_strength, "pcan_offset", pcan_offset, "gain_bits", gain_bits,
        "enable_log", enable_log, "scale_shift", scale_shift, "left_context",
        left_context, "right_context", right_context, "frame_stride",
        frame_stride, "zero_padding", zero_padding, "out_scale", out_scale,
        "out_type", out_type)
      return _result
    except _core._FallbackException:
      try:
        return audio_microfrontend_eager_fallback(
            audio, sample_rate=sample_rate, window_size=window_size,
            window_step=window_step, num_channels=num_channels,
            upper_band_limit=upper_band_limit,
            lower_band_limit=lower_band_limit, smoothing_bits=smoothing_bits,
            even_smoothing=even_smoothing, odd_smoothing=odd_smoothing,
            min_signal_remaining=min_signal_remaining,
            enable_pcan=enable_pcan, pcan_strength=pcan_strength,
            pcan_offset=pcan_offset, gain_bits=gain_bits,
            enable_log=enable_log, scale_shift=scale_shift,
            left_context=left_context, right_context=right_context,
            frame_stride=frame_stride, zero_padding=zero_padding,
            out_scale=out_scale, out_type=out_type, name=name, ctx=_ctx)
      except _core._SymbolicException:
        pass  # Add nodes to the TensorFlow graph.
      except (TypeError, ValueError):
        result = _dispatch.dispatch(
              audio_microfrontend, audio=audio, sample_rate=sample_rate,
                                   window_size=window_size,
                                   window_step=window_step,
                                   num_channels=num_channels,
                                   upper_band_limit=upper_band_limit,
                                   lower_band_limit=lower_band_limit,
                                   smoothing_bits=smoothing_bits,
                                   even_smoothing=even_smoothing,
                                   odd_smoothing=odd_smoothing,
                                   min_signal_remaining=min_signal_remaining,
                                   enable_pcan=enable_pcan,
                                   pcan_strength=pcan_strength,
                                   pcan_offset=pcan_offset,
                                   gain_bits=gain_bits, enable_log=enable_log,
                                   scale_shift=scale_shift,
                                   left_context=left_context,
                                   right_context=right_context,
                                   frame_stride=frame_stride,
                                   zero_padding=zero_padding,
                                   out_scale=out_scale, out_type=out_type,
                                   name=name)
        if result is not _dispatch.OpDispatcher.NOT_SUPPORTED:
          return result
        raise
    except _core._NotOkStatusException as e:
      if name is not None:
        message = e.message + " name: " + name
      else:
        message = e.message
      _six.raise_from(_core._status_to_exception(e.code, message), None)
  # Add nodes to the TensorFlow graph.
  if sample_rate is None:
    sample_rate = 16000
  sample_rate = _execute.make_int(sample_rate, "sample_rate")
  if window_size is None:
    window_size = 25
  window_size = _execute.make_int(window_size, "window_size")
  if window_step is None:
    window_step = 10
  window_step = _execute.make_int(window_step, "window_step")
  if num_channels is None:
    num_channels = 32
  num_channels = _execute.make_int(num_channels, "num_channels")
  if upper_band_limit is None:
    upper_band_limit = 7500
  upper_band_limit = _execute.make_float(upper_band_limit, "upper_band_limit")
  if lower_band_limit is None:
    lower_band_limit = 125
  lower_band_limit = _execute.make_float(lower_band_limit, "lower_band_limit")
  if smoothing_bits is None:
    smoothing_bits = 10
  smoothing_bits = _execute.make_int(smoothing_bits, "smoothing_bits")
  if even_smoothing is None:
    even_smoothing = 0.025
  even_smoothing = _execute.make_float(even_smoothing, "even_smoothing")
  if odd_smoothing is None:
    odd_smoothing = 0.06
  odd_smoothing = _execute.make_float(odd_smoothing, "odd_smoothing")
  if min_signal_remaining is None:
    min_signal_remaining = 0.05
  min_signal_remaining = _execute.make_float(min_signal_remaining, "min_signal_remaining")
  if enable_pcan is None:
    enable_pcan = False
  enable_pcan = _execute.make_bool(enable_pcan, "enable_pcan")
  if pcan_strength is None:
    pcan_strength = 0.95
  pcan_strength = _execute.make_float(pcan_strength, "pcan_strength")
  if pcan_offset is None:
    pcan_offset = 80
  pcan_offset = _execute.make_float(pcan_offset, "pcan_offset")
  if gain_bits is None:
    gain_bits = 21
  gain_bits = _execute.make_int(gain_bits, "gain_bits")
  if enable_log is None:
    enable_log = True
  enable_log = _execute.make_bool(enable_log, "enable_log")
  if scale_shift is None:
    scale_shift = 6
  scale_shift = _execute.make_int(scale_shift, "scale_shift")
  if left_context is None:
    left_context = 0
  left_context = _execute.make_int(left_context, "left_context")
  if right_context is None:
    right_context = 0
  right_context = _execute.make_int(right_context, "right_context")
  if frame_stride is None:
    frame_stride = 1
  frame_stride = _execute.make_int(frame_stride, "frame_stride")
  if zero_padding is None:
    zero_padding = False
  zero_padding = _execute.make_bool(zero_padding, "zero_padding")
  if out_scale is None:
    out_scale = 1
  out_scale = _execute.make_int(out_scale, "out_scale")
  if out_type is None:
    out_type = _dtypes.uint16
  out_type = _execute.make_type(out_type, "out_type")
  try:
    _, _, _op = _op_def_lib._apply_op_helper(
        "AudioMicrofrontend", audio=audio, sample_rate=sample_rate,
                              window_size=window_size,
                              window_step=window_step,
                              num_channels=num_channels,
                              upper_band_limit=upper_band_limit,
                              lower_band_limit=lower_band_limit,
                              smoothing_bits=smoothing_bits,
                              even_smoothing=even_smoothing,
                              odd_smoothing=odd_smoothing,
                              min_signal_remaining=min_signal_remaining,
                              enable_pcan=enable_pcan,
                              pcan_strength=pcan_strength,
                              pcan_offset=pcan_offset, gain_bits=gain_bits,
                              enable_log=enable_log, scale_shift=scale_shift,
                              left_context=left_context,
                              right_context=right_context,
                              frame_stride=frame_stride,
                              zero_padding=zero_padding, out_scale=out_scale,
                              out_type=out_type, name=name)
  except (TypeError, ValueError):
    result = _dispatch.dispatch(
          audio_microfrontend, audio=audio, sample_rate=sample_rate,
                               window_size=window_size,
                               window_step=window_step,
                               num_channels=num_channels,
                               upper_band_limit=upper_band_limit,
                               lower_band_limit=lower_band_limit,
                               smoothing_bits=smoothing_bits,
                               even_smoothing=even_smoothing,
                               odd_smoothing=odd_smoothing,
                               min_signal_remaining=min_signal_remaining,
                               enable_pcan=enable_pcan,
                               pcan_strength=pcan_strength,
                               pcan_offset=pcan_offset, gain_bits=gain_bits,
                               enable_log=enable_log, scale_shift=scale_shift,
                               left_context=left_context,
                               right_context=right_context,
                               frame_stride=frame_stride,
                               zero_padding=zero_padding, out_scale=out_scale,
                               out_type=out_type, name=name)
    if result is not _dispatch.OpDispatcher.NOT_SUPPORTED:
      return result
    raise
  _result = _op.outputs[:]
  _inputs_flat = _op.inputs
  _attrs = ("sample_rate", _op.get_attr("sample_rate"), "window_size",
            _op.get_attr("window_size"), "window_step",
            _op.get_attr("window_step"), "num_channels",
            _op.get_attr("num_channels"), "upper_band_limit",
            _op.get_attr("upper_band_limit"), "lower_band_limit",
            _op.get_attr("lower_band_limit"), "smoothing_bits",
            _op.get_attr("smoothing_bits"), "even_smoothing",
            _op.get_attr("even_smoothing"), "odd_smoothing",
            _op.get_attr("odd_smoothing"), "min_signal_remaining",
            _op.get_attr("min_signal_remaining"), "enable_pcan",
            _op.get_attr("enable_pcan"), "pcan_strength",
            _op.get_attr("pcan_strength"), "pcan_offset",
            _op.get_attr("pcan_offset"), "gain_bits",
            _op.get_attr("gain_bits"), "enable_log",
            _op.get_attr("enable_log"), "scale_shift",
            _op.get_attr("scale_shift"), "left_context",
            _op.get_attr("left_context"), "right_context",
            _op.get_attr("right_context"), "frame_stride",
            _op.get_attr("frame_stride"), "zero_padding",
            _op.get_attr("zero_padding"), "out_scale",
            _op.get_attr("out_scale"), "out_type",
            _op._get_attr_type("out_type"))
  _execute.record_gradient(
      "AudioMicrofrontend", _inputs_flat, _attrs, _result, name)
  _result, = _result
  return _result

def AudioMicrofrontend(audio, sample_rate=16000, window_size=25, window_step=10, num_channels=32, upper_band_limit=7500, lower_band_limit=125, smoothing_bits=10, even_smoothing=0.025, odd_smoothing=0.06, min_signal_remaining=0.05, enable_pcan=False, pcan_strength=0.95, pcan_offset=80, gain_bits=21, enable_log=True, scale_shift=6, left_context=0, right_context=0, frame_stride=1, zero_padding=False, out_scale=1, out_type=_dtypes.uint16, name=None):
  return audio_microfrontend(audio=audio, sample_rate=sample_rate, window_size=window_size, window_step=window_step, num_channels=num_channels, upper_band_limit=upper_band_limit, lower_band_limit=lower_band_limit, smoothing_bits=smoothing_bits, even_smoothing=even_smoothing, odd_smoothing=odd_smoothing, min_signal_remaining=min_signal_remaining, enable_pcan=enable_pcan, pcan_strength=pcan_strength, pcan_offset=pcan_offset, gain_bits=gain_bits, enable_log=enable_log, scale_shift=scale_shift, left_context=left_context, right_context=right_context, frame_stride=frame_stride, zero_padding=zero_padding, out_scale=out_scale, out_type=out_type, name=name)
AudioMicrofrontend.__doc__ = audio_microfrontend.__doc__
AudioMicrofrontend = _doc_controls.do_not_generate_docs(_kwarg_only(AudioMicrofrontend))
tf_export("raw_ops.AudioMicrofrontend")(AudioMicrofrontend)


def audio_microfrontend_eager_fallback(audio, sample_rate=16000, window_size=25, window_step=10, num_channels=32, upper_band_limit=7500, lower_band_limit=125, smoothing_bits=10, even_smoothing=0.025, odd_smoothing=0.06, min_signal_remaining=0.05, enable_pcan=False, pcan_strength=0.95, pcan_offset=80, gain_bits=21, enable_log=True, scale_shift=6, left_context=0, right_context=0, frame_stride=1, zero_padding=False, out_scale=1, out_type=_dtypes.uint16, name=None, ctx=None):
  r"""This is the slowpath function for Eager mode.
  This is for function audio_microfrontend
  """
  _ctx = ctx if ctx else _context.context()
  if sample_rate is None:
    sample_rate = 16000
  sample_rate = _execute.make_int(sample_rate, "sample_rate")
  if window_size is None:
    window_size = 25
  window_size = _execute.make_int(window_size, "window_size")
  if window_step is None:
    window_step = 10
  window_step = _execute.make_int(window_step, "window_step")
  if num_channels is None:
    num_channels = 32
  num_channels = _execute.make_int(num_channels, "num_channels")
  if upper_band_limit is None:
    upper_band_limit = 7500
  upper_band_limit = _execute.make_float(upper_band_limit, "upper_band_limit")
  if lower_band_limit is None:
    lower_band_limit = 125
  lower_band_limit = _execute.make_float(lower_band_limit, "lower_band_limit")
  if smoothing_bits is None:
    smoothing_bits = 10
  smoothing_bits = _execute.make_int(smoothing_bits, "smoothing_bits")
  if even_smoothing is None:
    even_smoothing = 0.025
  even_smoothing = _execute.make_float(even_smoothing, "even_smoothing")
  if odd_smoothing is None:
    odd_smoothing = 0.06
  odd_smoothing = _execute.make_float(odd_smoothing, "odd_smoothing")
  if min_signal_remaining is None:
    min_signal_remaining = 0.05
  min_signal_remaining = _execute.make_float(min_signal_remaining, "min_signal_remaining")
  if enable_pcan is None:
    enable_pcan = False
  enable_pcan = _execute.make_bool(enable_pcan, "enable_pcan")
  if pcan_strength is None:
    pcan_strength = 0.95
  pcan_strength = _execute.make_float(pcan_strength, "pcan_strength")
  if pcan_offset is None:
    pcan_offset = 80
  pcan_offset = _execute.make_float(pcan_offset, "pcan_offset")
  if gain_bits is None:
    gain_bits = 21
  gain_bits = _execute.make_int(gain_bits, "gain_bits")
  if enable_log is None:
    enable_log = True
  enable_log = _execute.make_bool(enable_log, "enable_log")
  if scale_shift is None:
    scale_shift = 6
  scale_shift = _execute.make_int(scale_shift, "scale_shift")
  if left_context is None:
    left_context = 0
  left_context = _execute.make_int(left_context, "left_context")
  if right_context is None:
    right_context = 0
  right_context = _execute.make_int(right_context, "right_context")
  if frame_stride is None:
    frame_stride = 1
  frame_stride = _execute.make_int(frame_stride, "frame_stride")
  if zero_padding is None:
    zero_padding = False
  zero_padding = _execute.make_bool(zero_padding, "zero_padding")
  if out_scale is None:
    out_scale = 1
  out_scale = _execute.make_int(out_scale, "out_scale")
  if out_type is None:
    out_type = _dtypes.uint16
  out_type = _execute.make_type(out_type, "out_type")
  audio = _ops.convert_to_tensor(audio, _dtypes.int16)
  _inputs_flat = [audio]
  _attrs = ("sample_rate", sample_rate, "window_size", window_size,
  "window_step", window_step, "num_channels", num_channels,
  "upper_band_limit", upper_band_limit, "lower_band_limit", lower_band_limit,
  "smoothing_bits", smoothing_bits, "even_smoothing", even_smoothing,
  "odd_smoothing", odd_smoothing, "min_signal_remaining",
  min_signal_remaining, "enable_pcan", enable_pcan, "pcan_strength",
  pcan_strength, "pcan_offset", pcan_offset, "gain_bits", gain_bits,
  "enable_log", enable_log, "scale_shift", scale_shift, "left_context",
  left_context, "right_context", right_context, "frame_stride", frame_stride,
  "zero_padding", zero_padding, "out_scale", out_scale, "out_type", out_type)
  _result = _execute.execute(b"AudioMicrofrontend", 1, inputs=_inputs_flat,
                             attrs=_attrs, ctx=_ctx, name=name)
  _execute.record_gradient(
      "AudioMicrofrontend", _inputs_flat, _attrs, _result, name)
  _result, = _result
  return _result

_ops.RegisterShape("AudioMicrofrontend")(None)

def _InitOpDefLibrary(op_list_proto_bytes):
  op_list = _op_def_pb2.OpList()
  op_list.ParseFromString(op_list_proto_bytes)
  _op_def_registry.register_op_list(op_list)
  op_def_lib = _op_def_library.OpDefLibrary()
  op_def_lib.add_op_list(op_list)
  return op_def_lib
# op {
#   name: "AudioMicrofrontend"
#   input_arg {
#     name: "audio"
#     type: DT_INT16
#   }
#   output_arg {
#     name: "filterbanks"
#     type_attr: "out_type"
#   }
#   attr {
#     name: "sample_rate"
#     type: "int"
#     default_value {
#       i: 16000
#     }
#   }
#   attr {
#     name: "window_size"
#     type: "int"
#     default_value {
#       i: 25
#     }
#   }
#   attr {
#     name: "window_step"
#     type: "int"
#     default_value {
#       i: 10
#     }
#   }
#   attr {
#     name: "num_channels"
#     type: "int"
#     default_value {
#       i: 32
#     }
#   }
#   attr {
#     name: "upper_band_limit"
#     type: "float"
#     default_value {
#       f: 7500
#     }
#   }
#   attr {
#     name: "lower_band_limit"
#     type: "float"
#     default_value {
#       f: 125
#     }
#   }
#   attr {
#     name: "smoothing_bits"
#     type: "int"
#     default_value {
#       i: 10
#     }
#   }
#   attr {
#     name: "even_smoothing"
#     type: "float"
#     default_value {
#       f: 0.025
#     }
#   }
#   attr {
#     name: "odd_smoothing"
#     type: "float"
#     default_value {
#       f: 0.06
#     }
#   }
#   attr {
#     name: "min_signal_remaining"
#     type: "float"
#     default_value {
#       f: 0.05
#     }
#   }
#   attr {
#     name: "enable_pcan"
#     type: "bool"
#     default_value {
#       b: false
#     }
#   }
#   attr {
#     name: "pcan_strength"
#     type: "float"
#     default_value {
#       f: 0.95
#     }
#   }
#   attr {
#     name: "pcan_offset"
#     type: "float"
#     default_value {
#       f: 80
#     }
#   }
#   attr {
#     name: "gain_bits"
#     type: "int"
#     default_value {
#       i: 21
#     }
#   }
#   attr {
#     name: "enable_log"
#     type: "bool"
#     default_value {
#       b: true
#     }
#   }
#   attr {
#     name: "scale_shift"
#     type: "int"
#     default_value {
#       i: 6
#     }
#   }
#   attr {
#     name: "left_context"
#     type: "int"
#     default_value {
#       i: 0
#     }
#   }
#   attr {
#     name: "right_context"
#     type: "int"
#     default_value {
#       i: 0
#     }
#   }
#   attr {
#     name: "frame_stride"
#     type: "int"
#     default_value {
#       i: 1
#     }
#   }
#   attr {
#     name: "zero_padding"
#     type: "bool"
#     default_value {
#       b: false
#     }
#   }
#   attr {
#     name: "out_scale"
#     type: "int"
#     default_value {
#       i: 1
#     }
#   }
#   attr {
#     name: "out_type"
#     type: "type"
#     default_value {
#       type: DT_UINT16
#     }
#     allowed_values {
#       list {
#         type: DT_UINT16
#         type: DT_FLOAT
#       }
#     }
#   }
# }
_op_def_lib = _InitOpDefLibrary(b"\n\223\005\n\022AudioMicrofrontend\022\t\n\005audio\030\005\032\027\n\013filterbanks\"\010out_type\"\027\n\013sample_rate\022\003int\032\003\030\200}\"\026\n\013window_size\022\003int\032\002\030\031\"\026\n\013window_step\022\003int\032\002\030\n\"\027\n\014num_channels\022\003int\032\002\030 \" \n\020upper_band_limit\022\005float\032\005%\000`\352E\" \n\020lower_band_limit\022\005float\032\005%\000\000\372B\"\031\n\016smoothing_bits\022\003int\032\002\030\n\"\036\n\016even_smoothing\022\005float\032\005%\315\314\314<\"\035\n\rodd_smoothing\022\005float\032\005%\217\302u=\"$\n\024min_signal_remaining\022\005float\032\005%\315\314L=\"\027\n\013enable_pcan\022\004bool\032\002(\000\"\035\n\rpcan_strength\022\005float\032\005%33s?\"\033\n\013pcan_offset\022\005float\032\005%\000\000\240B\"\024\n\tgain_bits\022\003int\032\002\030\025\"\026\n\nenable_log\022\004bool\032\002(\001\"\026\n\013scale_shift\022\003int\032\002\030\006\"\027\n\014left_context\022\003int\032\002\030\000\"\030\n\rright_context\022\003int\032\002\030\000\"\027\n\014frame_stride\022\003int\032\002\030\001\"\030\n\014zero_padding\022\004bool\032\002(\000\"\024\n\tout_scale\022\003int\032\002\030\001\"\034\n\010out_type\022\004type\032\0020\021:\006\n\0042\002\021\001")



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Core module for TensorFlow distribution objects and helpers.
"""

from __future__ import print_function as _print_function

import sys as _sys

from tensorflow.python.ops.distributions.bernoulli import Bernoulli
from tensorflow.python.ops.distributions.beta import Beta
from tensorflow.python.ops.distributions.categorical import Categorical
from tensorflow.python.ops.distributions.dirichlet import Dirichlet
from tensorflow.python.ops.distributions.dirichlet_multinomial import DirichletMultinomial
from tensorflow.python.ops.distributions.distribution import Distribution
from tensorflow.python.ops.distributions.distribution import FULLY_REPARAMETERIZED
from tensorflow.python.ops.distributions.distribution import NOT_REPARAMETERIZED
from tensorflow.python.ops.distributions.distribution import ReparameterizationType
from tensorflow.python.ops.distributions.exponential import Exponential
from tensorflow.python.ops.distributions.gamma import Gamma
from tensorflow.python.ops.distributions.kullback_leibler import RegisterKL
from tensorflow.python.ops.distributions.kullback_leibler import kl_divergence
from tensorflow.python.ops.distributions.laplace import Laplace
from tensorflow.python.ops.distributions.multinomial import Multinomial
from tensorflow.python.ops.distributions.normal import Normal
from tensorflow.python.ops.distributions.student_t import StudentT
from tensorflow.python.ops.distributions.uniform import Uniform

del _print_function

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "compat.v1.distributions", public_apis=None, deprecation=False,
      has_lite=False)



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Minimal runtime type checking library.

This module should not be considered public API.
"""
# TODO(ericmc,shoyer): Delete this in favor of using pytype or mypy
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import re

from tensorflow.python.util import tf_inspect
from tensorflow.python.util.compat import collections_abc

# used for register_type_abbreviation and _type_repr below.
_TYPE_ABBREVIATIONS = {}


class Type(object):
  """Base class for type checker types.

  The custom types defined in this module are based on types in the standard
  library's typing module (in Python 3.5):
  https://docs.python.org/3/library/typing.html

  The only difference should be that we use actual instances of Type classes to
  represent custom types rather than the metaclass magic typing uses to create
  new class objects. In practice, all this should mean is that we use
  `List(int)` rather than `List[int]`.

  Custom types should implement __instancecheck__ and inherit from Type. Every
  argument in the constructor must be a type or Type instance, and these
  arguments must be stored as a tuple on the `_types` attribute.
  """

  def __init__(self, *types):
    self._types = types

  def __repr__(self):
    args_repr = ", ".join(repr(t) for t in self._types)
    return "typecheck.%s(%s)" % (type(self).__name__, args_repr)


class _SingleArgumentType(Type):
  """Use this subclass for parametric types that accept only one argument."""

  def __init__(self, tpe):
    super(_SingleArgumentType, self).__init__(tpe)

  @property
  def _type(self):
    tpe, = self._types  # pylint: disable=unbalanced-tuple-unpacking
    return tpe


class _TwoArgumentType(Type):
  """Use this subclass for parametric types that accept two arguments."""

  def __init__(self, first_type, second_type):
    super(_TwoArgumentType, self).__init__(first_type, second_type)


class Union(Type):
  """A sum type.

  A correct type is any of the types provided.
  """

  def __instancecheck__(self, instance):
    return isinstance(instance, self._types)


class Optional(_SingleArgumentType):
  """An optional type.

  A correct type is either the provided type or NoneType.
  """

  def __instancecheck__(self, instance):
    # types.NoneType does not exist in Python 3
    return isinstance(instance, (self._type, type(None)))


class List(_SingleArgumentType):
  """A typed list.

  A correct type is a list where each element has the single provided type.
  """

  def __instancecheck__(self, instance):
    return (isinstance(instance, list) and
            all(isinstance(x, self._type) for x in instance))


class Sequence(_SingleArgumentType):
  """A typed sequence.

  A correct type is a sequence where each element has the single provided type.
  """

  def __instancecheck__(self, instance):
    return (isinstance(instance, collections_abc.Sequence) and
            all(isinstance(x, self._type) for x in instance))


class Collection(_SingleArgumentType):
  """A sized, iterable container.

  A correct type is an iterable and container with known size where each element
  has the single provided type.

  We use this in preference to Iterable because we check each instance of the
  iterable at runtime, and hence need to avoid iterables that could be
  exhausted.
  """

  def __instancecheck__(self, instance):
    return (isinstance(instance, collections_abc.Iterable) and
            isinstance(instance, collections_abc.Sized) and
            isinstance(instance, collections_abc.Container) and
            all(isinstance(x, self._type) for x in instance))


class Tuple(Type):
  """A typed tuple.

  A correct type is a tuple with the correct length where each element has
  the correct type.
  """

  def __instancecheck__(self, instance):
    return (isinstance(instance, tuple) and
            len(instance) == len(self._types) and
            all(isinstance(x, t) for x, t in zip(instance, self._types)))


class Mapping(_TwoArgumentType):
  """A typed mapping.

  A correct type has the correct parametric types for keys and values.
  """

  def __instancecheck__(self, instance):
    key_type, value_type = self._types  # pylint: disable=unbalanced-tuple-unpacking
    return (isinstance(instance, collections_abc.Mapping) and
            all(isinstance(k, key_type) for k in instance.keys()) and
            all(isinstance(k, value_type) for k in instance.values()))


class Dict(Mapping):
  """A typed dict.

  A correct type has the correct parametric types for keys and values.
  """

  def __instancecheck__(self, instance):
    return (isinstance(instance, dict) and
            super(Dict, self).__instancecheck__(instance))


def _replace_forward_references(t, context):
  """Replace forward references in the given type."""
  if isinstance(t, str):
    return context[t]
  elif isinstance(t, Type):
    return type(t)(*[_replace_forward_references(t, context) for t in t._types])  # pylint: disable=protected-access
  else:
    return t


def register_type_abbreviation(name, alias):
  """Register an abbreviation for a type in typecheck tracebacks.

  This makes otherwise very long typecheck errors much more readable.

  Example:
    typecheck.register_type_abbreviation(tf.compat.v1.Dimension,
    'tf.compat.v1.Dimension')

  Args:
    name: type or class to abbreviate.
    alias: string alias to substitute.
  """
  _TYPE_ABBREVIATIONS[name] = alias


def _type_repr(t):
  """A more succinct repr for typecheck tracebacks."""
  string = repr(t)
  for type_, alias in _TYPE_ABBREVIATIONS.items():
    string = string.replace(repr(type_), alias)
  string = re.sub(r"<(class|type) '([\w.]+)'>", r"\2", string)
  string = re.sub(r"typecheck\.(\w+)", r"\1", string)
  return string


class Error(TypeError):
  """Exception for typecheck failures."""


def accepts(*types):
  """A decorator which checks the input types of a function.

  Based on:
  http://stackoverflow.com/questions/15299878/how-to-use-python-decorators-to-check-function-arguments
  The above draws from:
  https://www.python.org/dev/peps/pep-0318/

  Args:
    *types: A list of Python types.

  Returns:
    A function to use as a decorator.
  """

  def check_accepts(f):
    """Check the types."""
    spec = tf_inspect.getargspec(f)

    num_function_arguments = len(spec.args)
    if len(types) != num_function_arguments:
      raise Error(
          "Function %r has %d arguments but only %d types were provided in the "
          "annotation." % (f, num_function_arguments, len(types)))

    if spec.defaults:
      num_defaults = len(spec.defaults)
      for (name, a, t) in zip(spec.args[-num_defaults:], spec.defaults,
                              types[-num_defaults:]):
        allowed_type = _replace_forward_references(t, f.__globals__)
        if not isinstance(a, allowed_type):
          raise Error("default argument value %r of type %r is not an instance "
                      "of the allowed type %s for the %s argument to %r" %
                      (a, type(a), _type_repr(allowed_type), name, f))

    @functools.wraps(f)
    def new_f(*args, **kwds):
      """A helper function."""
      for (a, t) in zip(args, types):
        allowed_type = _replace_forward_references(t, f.__globals__)
        if not isinstance(a, allowed_type):
          raise Error("%r of type %r is not an instance of the allowed type %s "
                      "for %r" % (a, type(a), _type_repr(allowed_type), f))
      return f(*args, **kwds)

    return new_f

  return check_accepts


def returns(*types):
  """A decorator which checks the return types of a function.

  Based on:
  http://stackoverflow.com/questions/15299878/how-to-use-python-decorators-to-check-function-arguments
  The above draws from:
  https://www.python.org/dev/peps/pep-0318/

  Args:
    *types: A list of Python types. A list of one element corresponds to a
      single return value. A list of several elements corresponds to several
      return values. Note that a function with no explicit return value has an
      implicit NoneType return and should be annotated correspondingly.

  Returns:
    A function to use as a decorator.
  """

  def check_returns(f):
    """Check the types."""
    if not types:
      raise TypeError("A return type annotation must contain at least one type")

    @functools.wraps(f)
    def new_f(*args, **kwds):
      """A helper function."""
      return_value = f(*args, **kwds)

      if len(types) == 1:
        # The function has a single return value.
        allowed_type = _replace_forward_references(types[0], f.__globals__)
        if not isinstance(return_value, allowed_type):
          raise Error(
              "%r of type %r is not an instance of the allowed type %s "
              "for %r" %
              (return_value, type(return_value), _type_repr(allowed_type), f))

      else:
        if len(return_value) != len(types):
          raise Error("Function %r has %d return values but only %d types were "
                      "provided in the annotation." %
                      (f, len(return_value), len(types)))

        for (r, t) in zip(return_value, types):
          allowed_type = _replace_forward_references(t, f.__globals__)
          if not isinstance(r, allowed_type):
            raise Error("%r of type %r is not an instance of allowed type %s "
                        "for %r" % (r, type(r), _type_repr(allowed_type), f))

      return return_value

    return new_f

  return check_returns



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Model pruning implementation in tensorflow."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=unused-import
from tensorflow.contrib.model_pruning.python.layers.layers import masked_conv2d
from tensorflow.contrib.model_pruning.python.layers.layers import masked_convolution
from tensorflow.contrib.model_pruning.python.layers.layers import masked_fully_connected
from tensorflow.contrib.model_pruning.python.layers.rnn_cells import MaskedBasicLSTMCell
from tensorflow.contrib.model_pruning.python.layers.rnn_cells import MaskedLSTMCell
from tensorflow.contrib.model_pruning.python.learning import train
from tensorflow.contrib.model_pruning.python.pruning import apply_mask
from tensorflow.contrib.model_pruning.python.pruning import get_masked_weights
from tensorflow.contrib.model_pruning.python.pruning import get_masks
from tensorflow.contrib.model_pruning.python.pruning import get_pruning_hparams
from tensorflow.contrib.model_pruning.python.pruning import get_thresholds
from tensorflow.contrib.model_pruning.python.pruning import get_weight_sparsity
from tensorflow.contrib.model_pruning.python.pruning import get_weights
from tensorflow.contrib.model_pruning.python.pruning import Pruning
from tensorflow.contrib.model_pruning.python.strip_pruning_vars_lib import graph_def_from_checkpoint
from tensorflow.contrib.model_pruning.python.strip_pruning_vars_lib import strip_pruning_vars_fn

# pylint: enable=unused-import

from tensorflow.python.util.all_util import remove_undocumented

_allowed_symbols = [
    'masked_convolution', 'masked_conv2d', 'masked_fully_connected',
    'MaskedBasicLSTMCell', 'MaskedLSTMCell', 'train', 'apply_mask',
    'get_masked_weights', 'get_masks', 'get_pruning_hparams', 'get_thresholds',
    'get_weights', 'get_weight_sparsity', 'Pruning', 'strip_pruning_vars_fn',
    'graph_def_from_checkpoint'
]

remove_undocumented(__name__, _allowed_symbols)
